@@ -5,8 +5,9 @@ import sys
 import lazy_import
 from smac.utils import dependencies
 
-__version__ = '0.11.0'
-__author__ = 'Marius Lindauer, Matthias Feurer, Katharina Eggensperger, Joshua Marben, André Biedenkapp, Aaron Klein, Stefan Falkner and Frank Hutter'
+__version__ = '0.12.0'
+__author__ = 'Marius Lindauer, Matthias Feurer, Katharina Eggensperger, Joshua Marben, André Biedenkapp, Aaron Klein,'\
+    'Stefan Falkner and Frank Hutter'
 
 
 with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as fh:
@@ -17,10 +18,12 @@ with open(os.path.join(os.path.dirname(__file__), 'extras_require.json')) as fh:
 
 extras_installed = set()
 for name, requirements in extras_require.items():
+    if name in ['documentation', 'test']:
+        continue
     if dependencies.are_valid_packages(requirements):
         extras_installed.add(name)
     for requirement in requirements:
-        package_name = dependencies.RE_PATTERN.match(requirement).group('name')
+        package_name = dependencies.RE_PATTERN.match(requirement).group('name')  # type: ignore[union-attr] # noqa F821
         if package_name == 'scikit-optimize':
             package_name = 'skopt'
         lazy_import.lazy_module(package_name)
