@@ -2,17 +2,23 @@ import logging
 import typing
 
 import numpy as np
-from lazy_import import lazy_callable
 from scipy import optimize
 
 from smac.configspace import ConfigurationSpace
 from smac.epm.base_gp import BaseModel
+from smac.epm.gp_base_prior import Prior
 from smac.utils.constants import VERY_SMALL_NUMBER
 
+if typing.TYPE_CHECKING:
+    from skopt.learning.gaussian_process.kernels import Kernel
+    from skopt.learning.gaussian_process import GaussianProcessRegressor
+else:
+    from lazy_import import lazy_callable
+    Kernel = lazy_callable('skopt.learning.gaussian_process.kernels.Kernel')
+    GaussianProcessRegressor = lazy_callable(
+        'skopt.learning.gaussian_process.GaussianProcessRegressor')
+
 logger = logging.getLogger(__name__)
-Kernel = lazy_callable('skopt.learning.gaussian_process.kernels.Kernel')
-GaussianProcessRegressor = lazy_callable(
-    'skopt.learning.gaussian_process.GaussianProcessRegressor')
 
 
 class GaussianProcess(BaseModel):
@@ -29,14 +35,14 @@ class GaussianProcess(BaseModel):
 
     Parameters
     ----------
-    types : np.ndarray (D)
+    types : List[int]
         Specifies the number of categorical values of an input dimension where
         the i-th entry corresponds to the i-th input dimension. Let's say we
         have 2 dimension where the first dimension consists of 3 different
         categorical choices and the second dimension is continuous than we
-        have to pass np.array([2, 0]). Note that we count starting from 0.
-    bounds : list
-        Specifies the bounds for continuous features.
+        have to pass [3, 0]. Note that we count starting from 0.
+    bounds : List[Tuple[float, float]]
+        bounds of input dimensions: (lower, uppper) for continuous dims; (n_cat, np.nan) for categorical dims
     seed : int
         Model seed.
     kernel : george kernel object
@@ -46,36 +52,47 @@ class GaussianProcess(BaseModel):
         it implements the Prior interface.
     normalize_y : bool
         Zero mean unit variance normalization of the output values
-    rng: np.random.RandomState
-        Random number generator
+    n_opt_restart : int
+        Number of restarts for GP hyperparameter optimization
+    instance_features : np.ndarray (I, K)
+        Contains the K dimensional instance features of the I different instances
+    pca_components : float
+        Number of components to keep when using PCA to reduce dimensionality of instance features. Requires to
+        set n_feats (> pca_dims).
     """
 
     def __init__(
         self,
         configspace: ConfigurationSpace,
-        types: np.ndarray,
+        types: typing.List[int],
         bounds: typing.List[typing.Tuple[float, float]],
         seed: int,
         kernel: Kernel,
-        normalize_y: bool=True,
-        n_opt_restarts=10,
-        **kwargs
+        normalize_y: bool = True,
+        n_opt_restarts: int = 10,
+        instance_features: typing.Optional[np.ndarray] = None,
+        pca_components: typing.Optional[int] = None,
     ):
+        super().__init__(
+            configspace=configspace,
+            types=types,
+            bounds=bounds,
+            seed=seed,
+            kernel=kernel,
+            instance_features=instance_features,
+            pca_components=pca_components,
+        )
 
-        super().__init__(configspace=configspace, types=types, bounds=bounds, seed=seed, **kwargs)
-
-        self.kernel = kernel
-        self.gp = None
         self.normalize_y = normalize_y
         self.n_opt_restarts = n_opt_restarts
 
-        self.hypers = []
+        self.hypers = np.empty((0, ))
         self.is_trained = False
         self._n_ll_evals = 0
 
         self._set_has_conditions()
 
-    def _train(self, X: np.ndarray, y: np.ndarray, do_optimize: bool=True):
+    def _train(self, X: np.ndarray, y: np.ndarray, do_optimize: bool = True) -> 'GaussianProcess':
         """
         Computes the Cholesky decomposition of the covariance of X and
         estimates the GP hyperparameters by optimizing the marginal
@@ -97,19 +114,17 @@ class GaussianProcess(BaseModel):
         X = self._impute_inactive(X)
         if self.normalize_y:
             y = self._normalize_y(y)
+        if len(y.shape) == 1:
+            self.n_objectives_ = 1
+        else:
+            self.n_objectives_ = y.shape[1]
+        if self.n_objectives_ == 1:
+            y = y.flatten()
 
         n_tries = 10
         for i in range(n_tries):
             try:
-                self.gp = GaussianProcessRegressor(
-                    kernel=self.kernel,
-                    normalize_y=False,
-                    optimizer=None,
-                    n_restarts_optimizer=-1,  # Do not use scikit-learn's optimization routine
-                    alpha=0,  # Governed by the kernel
-                    noise=None,
-                    random_state=self.rng,
-                )
+                self.gp = self._get_gp()
                 self.gp.fit(X, y)
                 break
             except np.linalg.LinAlgError as e:
@@ -129,6 +144,18 @@ class GaussianProcess(BaseModel):
             self.hypers = self.gp.kernel.theta
 
         self.is_trained = True
+        return self
+
+    def _get_gp(self) -> GaussianProcessRegressor:
+        return GaussianProcessRegressor(
+            kernel=self.kernel,
+            normalize_y=False,
+            optimizer=None,
+            n_restarts_optimizer=-1,  # Do not use scikit-learn's optimization routine
+            alpha=0,  # Governed by the kernel
+            noise=None,
+            random_state=self.rng,
+        )
 
     def _nll(self, theta: np.ndarray) -> typing.Tuple[float, np.ndarray]:
         """
@@ -182,6 +209,8 @@ class GaussianProcess(BaseModel):
         p0 = [self.gp.kernel.theta]
         if self.n_opt_restarts > 0:
             dim_samples = []
+
+            prior = None  # type: typing.Optional[typing.Union[typing.List[Prior], Prior]]
             for dim, hp_bound in enumerate(log_bounds):
                 prior = self._all_priors[dim]
                 # Always sample from the first prior
@@ -190,6 +219,7 @@ class GaussianProcess(BaseModel):
                         prior = None
                     else:
                         prior = prior[0]
+                prior = typing.cast(typing.Optional[Prior], prior)
                 if prior is None:
                     try:
                         sample = self.rng.uniform(
@@ -213,7 +243,7 @@ class GaussianProcess(BaseModel):
                 theta_star = theta
         return theta_star
 
-    def _predict(self, X_test: np.ndarray, full_cov: bool=False):
+    def _predict(self, X_test: np.ndarray, full_cov: bool = False) -> typing.Tuple[np.ndarray, np.ndarray]:
         r"""
         Returns the predictive mean and variance of the objective function at
         the given test points.
@@ -250,7 +280,7 @@ class GaussianProcess(BaseModel):
 
         return mu, var
 
-    def sample_functions(self, X_test: np.ndarray, n_funcs: int=1) -> np.ndarray:
+    def sample_functions(self, X_test: np.ndarray, n_funcs: int = 1) -> np.ndarray:
         """
         Samples F function values from the current posterior at the N
         specified test points.
@@ -273,7 +303,6 @@ class GaussianProcess(BaseModel):
 
         X_test = self._impute_inactive(X_test)
         funcs = self.gp.sample_y(X_test, n_samples=n_funcs, random_state=self.rng)
-        funcs = np.squeeze(funcs, axis=1)
 
         if self.normalize_y:
             funcs = self._untransform_y(funcs)
