@@ -17,14 +17,29 @@ __email__ = "lindauer@cs.uni-freiburg.de"
 __version__ = "0.0.1"
 
 
-RunKey = collections.namedtuple(
-    'RunKey', ['config_id', 'instance_id', 'seed'])
+# NOTE class instead of collection to have a default value for budget in RunKey
+class RunKey(collections.namedtuple('RunKey', ['config_id', 'instance_id', 'seed', 'budget'])):
+    __slots__ = ()
+
+    def __new__(
+        cls,  # No type annotation because the 1st argument for a namedtuble is always the class type,
+              # see https://docs.python.org/3/reference/datamodel.html#object.__new__
+        config_id: int,
+        instance_id: typing.Optional[str],
+        seed: typing.Optional[int],
+        budget: float = 0.0,
+    ) -> 'RunKey':
+        return super().__new__(cls, config_id, instance_id, seed, budget)
 
 
 InstSeedKey = collections.namedtuple(
     'InstSeedKey', ['instance', 'seed'])
 
 
+InstSeedBudgetKey = collections.namedtuple(
+    'InstSeedBudgetKey', ['instance', 'seed', 'budget'])
+
+
 RunValue = collections.namedtuple(
     'RunValue', ['cost', 'time', 'status', 'additional_info'])
 
@@ -36,7 +51,7 @@ class EnumEncoder(json.JSONEncoder):
     to deserialize from json.
     """
 
-    def default(self, obj):
+    def default(self, obj: object) -> typing.Any:
         if isinstance(obj, StatusType):
             return {"__enum__": str(obj)}
         return json.JSONEncoder.default(self, obj)
@@ -68,7 +83,18 @@ class RunHistory(object):
 
     """Container for target algorithm run information.
 
-    **Note:** Guaranteed to be picklable.
+    Most importantly, the runhistory contains an efficient mapping from each evaluated configuration to the
+    empirical cost observed on either the full instance set or a subset. The cost is the average over all
+    observed costs for one configuration:
+
+    * If using budgets for a single instance, only the cost on the highest observed budget is returned.
+    * If using instances as the budget, the average cost over all evaluated instances is returned.
+    * Theoretically, the runhistory object can handle instances and budgets at the same time. This is
+      neither used nor tested.
+
+    Note
+    ----
+    Guaranteed to be picklable.
 
     Attributes
     ----------
@@ -78,26 +104,24 @@ class RunHistory(object):
         Maps config -> id
     ids_config : dict
         Maps id -> config
-    cost_per_config : dict
-        Maps config_id -> cost
-    runs_per_config : dict
+    num_runs_per_config : dict
         Maps config_id -> number of runs
 
-    aggregate_func
-    overwrite_existing_runs
+    Parameters
+    ----------
+    overwrite_existing_runs : bool (default=True)
+        If set to ``True`` and a run of a configuration on an instance-budget-seed-pair already exists,
+        it is overwritten.
     """
 
     def __init__(
         self,
-        aggregate_func: typing.Callable,
-        overwrite_existing_runs: bool=False
+        overwrite_existing_runs: bool = False
     ) -> None:
         """Constructor
 
         Parameters
         ----------
-        aggregate_func: callable
-            function to aggregate perf across instances
         overwrite_existing_runs: bool
             allows to overwrites old results if pairs of
             algorithm-instance-seed were measured
@@ -114,30 +138,36 @@ class RunHistory(object):
 
         # for fast access, we have also an unordered data structure
         # to get all instance seed pairs of a configuration
-        self._configid_to_inst_seed = {}  # type: typing.Dict[int, InstSeedKey]
+        self._configid_to_inst_seed_budget = {}  # type: typing.Dict[int, typing.Dict[InstSeedKey, typing.List[float]]]
 
         self.config_ids = {}  # type: typing.Dict[Configuration, int]
         self.ids_config = {}  # type: typing.Dict[int, Configuration]
         self._n_id = 0
 
         # Stores cost for each configuration ID
-        self.cost_per_config = {}  # type: typing.Dict[int, float]
+        self._cost_per_config = {}  # type: typing.Dict[int, float]
         # runs_per_config maps the configuration ID to the number of runs for that configuration
         # and is necessary for computing the moving average
-        self.runs_per_config = {}  # type: typing.Dict[int, int]
+        self.num_runs_per_config = {}  # type: typing.Dict[int, int]
 
         # Store whether a datapoint is "external", which means it was read from
         # a JSON file. Can be chosen to not be written to disk
         self.external = {}  # type: typing.Dict[RunKey, DataOrigin]
 
-        self.aggregate_func = aggregate_func
         self.overwrite_existing_runs = overwrite_existing_runs
 
-    def add(self, config: Configuration, cost: float, time: float,
-            status: StatusType, instance_id: str=None,
-            seed: int=None,
-            additional_info: dict=None,
-            origin: DataOrigin=DataOrigin.INTERNAL):
+    def add(
+        self,
+        config: Configuration,
+        cost: float,
+        time: float,
+        status: StatusType,
+        instance_id: typing.Optional[str] = None,
+        seed: typing.Optional[int] = None,
+        budget: float = 0,
+        additional_info: typing.Optional[typing.Dict] = None,
+        origin: DataOrigin = DataOrigin.INTERNAL,
+    ) -> None:
         """Adds a data of a new target algorithm (TA) run;
         it will update data if the same key values are used
         (config, instance_id, seed)
@@ -156,6 +186,8 @@ class RunHistory(object):
                 String representing an instance (default: None)
             seed: int
                 Random seed used by TA (default: None)
+            budget: float
+                budget (cutoff) used in intensifier to limit TA (default: 0)
             additional_info: dict
                 Additional run infos (could include further returned
                 information from TA or fields such as start time and host_id)
@@ -163,14 +195,18 @@ class RunHistory(object):
                 Defines how data will be used.
         """
 
-        config_id = self.config_ids.get(config)
-        if config_id is None:
+        # Get the config id
+        config_id_tmp = self.config_ids.get(config)
+        if config_id_tmp is None:
             self._n_id += 1
             self.config_ids[config] = self._n_id
-            config_id = self.config_ids.get(config)
+            config_id = typing.cast(int, self.config_ids.get(config))
             self.ids_config[self._n_id] = config
+        else:
+            config_id = typing.cast(int, config_id_tmp)
 
-        k = RunKey(config_id, instance_id, seed)
+        # Construct keys and values for the data dictionary
+        k = RunKey(config_id, instance_id, seed, budget)
         v = RunValue(cost, time, status, additional_info)
 
         # Each runkey is supposed to be used only once. Repeated tries to add
@@ -185,7 +221,7 @@ class RunHistory(object):
             self._add(k, v, status, origin)
 
     def _add(self, k: RunKey, v: RunValue, status: StatusType,
-             origin: DataOrigin):
+             origin: DataOrigin) -> None:
         """Actual function to add new entry to data structures
 
         TODO
@@ -195,60 +231,46 @@ class RunHistory(object):
         self.external[k] = origin
 
         if origin in (DataOrigin.INTERNAL, DataOrigin.EXTERNAL_SAME_INSTANCES) \
-                and status != StatusType.CAPPED:
+                and status != StatusType.CAPPED:  # Capped data is added above
             # also add to fast data structure
             is_k = InstSeedKey(k.instance_id, k.seed)
-            self._configid_to_inst_seed[
-                k.config_id] = self._configid_to_inst_seed.get(k.config_id, [])
-            if is_k not in self._configid_to_inst_seed[k.config_id]:
-                self._configid_to_inst_seed[k.config_id].append(is_k)
+            self._configid_to_inst_seed_budget[k.config_id] = self._configid_to_inst_seed_budget.get(k.config_id, {})
+            if is_k not in self._configid_to_inst_seed_budget[k.config_id].keys():
+                # add new inst-seed-key with budget to main dict
+                self._configid_to_inst_seed_budget[k.config_id][is_k] = [k.budget]
+            elif k.budget not in is_k:
+                # append new budget to existing inst-seed-key dict
+                self._configid_to_inst_seed_budget[k.config_id][is_k].append(k.budget)
 
-            if not self.overwrite_existing_runs:
-                # assumes an average across runs as cost function aggregation
+            # if budget is used, then update cost instead of incremental updates
+            if not self.overwrite_existing_runs and k.budget == 0:
+                # assumes an average across runs as cost function aggregation, this is used for algorithm configuration
+                # (incremental updates are used to save time as getting the cost for > 100 instances is high)
                 self.incremental_update_cost(self.ids_config[k.config_id], v.cost)
             else:
+                # this is when budget > 0 (only successive halving and hyperband so far)
                 self.update_cost(config=self.ids_config[k.config_id])
+                if k.budget > 0:
+                    if self.num_runs_per_config[k.config_id] != 1:  # This is updated in update_cost
+                        raise ValueError('This should not happen!')
 
-    def update_cost(self, config: Configuration):
+    def update_cost(self, config: Configuration) -> None:
         """Store the performance of a configuration across the instances in
-        self.cost_perf_config and also updates self.runs_per_config;
-        uses self.aggregate_func
+        self.cost_per_config and also updates self.runs_per_config;
 
         Parameters
         ----------
         config: Configuration
             configuration to update cost based on all runs in runhistory
         """
-        inst_seeds = set(self.get_runs_for_config(config))
-        perf = self.aggregate_func(config, self, inst_seeds)
+        # removing duplicates while keeping the order
+        inst_seed_budgets = list(dict.fromkeys(self.get_runs_for_config(config, only_max_observed_budget=True)))
+        perf = self.average_cost(config, inst_seed_budgets)
         config_id = self.config_ids[config]
-        self.cost_per_config[config_id] = perf
-        self.runs_per_config[config_id] = len(inst_seeds)
+        self._cost_per_config[config_id] = perf
+        self.num_runs_per_config[config_id] = len(inst_seed_budgets)
 
-    def compute_all_costs(self, instances: typing.List[str]=None):
-        """Computes the cost of all configurations from scratch and overwrites
-        self.cost_perf_config and self.runs_per_config accordingly;
-
-        Parameters
-        ----------
-        instances: typing.List[str]
-            list of instances; if given, cost is only computed wrt to this instance set
-        """
-
-        self.cost_per_config = {}
-        self.runs_per_config = {}
-        for config, config_id in self.config_ids.items():
-            inst_seeds = set(self.get_runs_for_config(config))
-            if instances is not None:
-                inst_seeds = list(
-                    filter(lambda x: x.instance in instances, inst_seeds))
-
-            if inst_seeds:  # can be empty if never saw any runs on <instances>
-                perf = self.aggregate_func(config, self, inst_seeds)
-                self.cost_per_config[config_id] = perf
-                self.runs_per_config[config_id] = len(inst_seeds)
-
-    def incremental_update_cost(self, config: Configuration, cost: float):
+    def incremental_update_cost(self, config: Configuration, cost: float) -> None:
         """Incrementally updates the performance of a configuration by using a
         moving average;
 
@@ -261,14 +283,15 @@ class RunHistory(object):
         """
 
         config_id = self.config_ids[config]
-        n_runs = self.runs_per_config.get(config_id, 0)
-        old_cost = self.cost_per_config.get(config_id, 0.)
-        self.cost_per_config[config_id] = (
-            (old_cost * n_runs) + cost) / (n_runs + 1)
-        self.runs_per_config[config_id] = n_runs + 1
+        n_runs = self.num_runs_per_config.get(config_id, 0)
+        old_cost = self._cost_per_config.get(config_id, 0.)
+        self._cost_per_config[config_id] = ((old_cost * n_runs) + cost) / (n_runs + 1)
+        self.num_runs_per_config[config_id] = n_runs + 1
 
-    def get_cost(self, config: Configuration):
-        """Returns empirical cost for a configuration; uses  self.cost_per_config
+    def get_cost(self, config: Configuration) -> float:
+        """Returns empirical cost for a configuration.
+
+        See the class docstring for how the costs are computed. The costs are not re-computed, but are read from cache.
 
         Parameters
         ----------
@@ -279,50 +302,36 @@ class RunHistory(object):
         cost: float
             Computed cost for configuration
         """
-        config_id = self.config_ids[config]
-        return self.cost_per_config.get(config_id, np.nan)
+        config_id = self.config_ids.get(config)
+        return self._cost_per_config.get(config_id, np.nan)  # type: ignore[arg-type] # noqa F821
 
-    def get_runs_for_config(self, config: Configuration):
+    def get_runs_for_config(self,
+                            config: Configuration, only_max_observed_budget: bool) -> typing.List[InstSeedBudgetKey]:
         """Return all runs (instance seed pairs) for a configuration.
 
         Parameters
         ----------
         config : Configuration from ConfigSpace
             Parameter configuration
-
+        only_max_observed_budget : bool
+            Select only the maximally observed budget run for this configuration
         Returns
         -------
-        instance_seed_pairs : list<tuples of instance, seed>
+        instance_seed_budget_pairs : list<tuples of instance, seed, budget>
         """
         config_id = self.config_ids.get(config)
-        return self._configid_to_inst_seed.get(config_id, [])
+        runs = self._configid_to_inst_seed_budget.get(config_id, {}).copy()  # type: ignore[arg-type] # noqa F821
 
-    def get_instance_costs_for_config(self, config: Configuration):
-        """
-            Returns the average cost per instance (across seeds)
-            for a configuration
-            Parameters
-            ----------
-            config : Configuration from ConfigSpace
-                Parameter configuration
+        # select only the max budget run if specified
+        if only_max_observed_budget:
+            for k, v in runs.items():
+                runs[k] = [max(v)]
 
-            Returns
-            -------
-            cost_per_inst: dict<instance name<str>, cost<float>>
-        """
-        config_id = self.config_ids.get(config)
-        runs_ = self._configid_to_inst_seed.get(config_id, [])
-        cost_per_inst = {}
-        for inst, seed in runs_:
-            cost_per_inst[inst] = cost_per_inst.get(inst,[])
-            rkey = RunKey(config_id, inst, seed)
-            vkey = self.data[rkey]
-            cost_per_inst[inst].append(vkey.cost)
-        cost_per_inst = dict([(inst, np.mean(costs)) for inst, costs in cost_per_inst.items()])
-        return cost_per_inst
+        # convert to inst-seed-budget key
+        rval = [InstSeedBudgetKey(k.instance, k.seed, budget) for k, v in runs.items() for budget in v]
+        return rval
 
-
-    def get_all_configs(self):
+    def get_all_configs(self) -> typing.List[Configuration]:
         """Return all configurations in this RunHistory object
 
         Returns
@@ -331,7 +340,30 @@ class RunHistory(object):
         """
         return list(self.config_ids.keys())
 
-    def empty(self):
+    def get_all_configs_per_budget(
+        self,
+        budget_subset: typing.Optional[typing.List] = None,
+    ) -> typing.List[Configuration]:
+        """
+        Return all configs in this RunHistory object that have been run on one of these budgets
+
+        Parameter
+        ---------
+            budget_subset: list
+
+        Returns
+        -------
+            parameter configurations: list
+        """
+        if budget_subset is None:
+            return self.get_all_configs()
+        configs = []
+        for c, i, s, b in self.data.keys():
+            if b in budget_subset:
+                configs.append(self.ids_config[c])
+        return configs
+
+    def empty(self) -> bool:
         """Check whether or not the RunHistory is empty.
 
         Returns
@@ -342,7 +374,7 @@ class RunHistory(object):
         """
         return len(self.data) == 0
 
-    def save_json(self, fn: str="runhistory.json", save_external: bool=False):
+    def save_json(self, fn: str = "runhistory.json", save_external: bool = False) -> None:
         """
         saves runhistory on disk
 
@@ -355,7 +387,8 @@ class RunHistory(object):
         """
         data = [([int(k.config_id),
                   str(k.instance_id) if k.instance_id is not None else None,
-                  int(k.seed)], list(v))
+                  int(k.seed),
+                  float(k.budget) if k[3] is not None else 0], list(v))
                 for k, v in self.data.items()
                 if save_external or self.external[k] == DataOrigin.INTERNAL]
         config_ids_to_serialize = set([entry[0][0] for entry in data])
@@ -364,15 +397,14 @@ class RunHistory(object):
                    if id_ in config_ids_to_serialize}
         config_origins = {id_: conf.origin
                           for id_, conf in self.ids_config.items()
-                          if (id_ in config_ids_to_serialize and
-                              conf.origin is not None)}
+                          if (id_ in config_ids_to_serialize and conf.origin is not None)}
 
         with open(fn, "w") as fp:
             json.dump({"data": data,
                        "config_origins": config_origins,
                        "configs": configs}, fp, cls=EnumEncoder, indent=2)
 
-    def load_json(self, fn: str, cs: ConfigurationSpace):
+    def load_json(self, fn: str, cs: ConfigurationSpace) -> None:
         """Load and runhistory in json representation from disk.
 
         Overwrites current runhistory!
@@ -416,10 +448,15 @@ class RunHistory(object):
                      status=StatusType(v[2]),
                      instance_id=k[1],
                      seed=int(k[2]),
+                     budget=float(k[3]) if len(k) == 4 else 0,
                      additional_info=v[3])
 
-    def update_from_json(self, fn: str, cs: ConfigurationSpace,
-                         origin: DataOrigin=DataOrigin.EXTERNAL_SAME_INSTANCES):
+    def update_from_json(
+        self,
+        fn: str,
+        cs: ConfigurationSpace,
+        origin: DataOrigin = DataOrigin.EXTERNAL_SAME_INSTANCES,
+    ) -> None:
         """Update the current runhistory by adding new runs from a json file.
 
         Parameters
@@ -431,12 +468,15 @@ class RunHistory(object):
         origin : DataOrigin
             What to store as data origin.
         """
-        new_runhistory = RunHistory(self.aggregate_func)
+        new_runhistory = RunHistory()
         new_runhistory.load_json(fn, cs)
         self.update(runhistory=new_runhistory, origin=origin)
 
-    def update(self, runhistory: 'RunHistory',
-               origin: DataOrigin=DataOrigin.EXTERNAL_SAME_INSTANCES):
+    def update(
+        self,
+        runhistory: 'RunHistory',
+        origin: DataOrigin = DataOrigin.EXTERNAL_SAME_INSTANCES,
+    ) -> None:
         """Update the current runhistory by adding new runs from a RunHistory.
 
         Parameters
@@ -445,7 +485,7 @@ class RunHistory(object):
             Runhistory with additional data to be added to self
         origin: DataOrigin
             If set to ``INTERNAL`` or ``EXTERNAL_FULL`` the data will be
-            added to the internal data structure self._configid_to_inst_seed
+            added to the internal data structure self._configid_to_inst_seed_budget
             and be available :meth:`through get_runs_for_config`.
         """
 
@@ -454,10 +494,150 @@ class RunHistory(object):
         # correctly by assigning an ID to unknown configurations and re-using
         #  the ID
         for key, value in runhistory.data.items():
-            config_id, instance_id, seed = key
+            config_id, instance_id, seed, budget = key
             cost, time, status, additional_info = value
             config = runhistory.ids_config[config_id]
             self.add(config=config, cost=cost, time=time,
                      status=status, instance_id=instance_id,
-                     seed=seed, additional_info=additional_info,
+                     seed=seed, budget=budget, additional_info=additional_info,
                      origin=origin)
+
+    def _cost(
+        self,
+        config: Configuration,
+        instance_seed_budget_keys: typing.Optional[typing.Iterable[InstSeedBudgetKey]] = None,
+    ) -> typing.List[float]:
+        """Return array of all costs for the given config for further calculations.
+
+        Parameters
+        ----------
+        config : Configuration
+            Configuration to calculate objective for
+        instance_seed_budget_keys : list, optional (default=None)
+            List of tuples of instance-seeds-budget keys. If None, the run_history is
+            queried for all runs of the given configuration.
+
+        Returns
+        -------
+        Costs: list
+            Array of all costs
+        """
+        try:
+            id_ = self.config_ids[config]
+        except KeyError:  # challenger was not running so far
+            return []
+
+        if instance_seed_budget_keys is None:
+            instance_seed_budget_keys = self.get_runs_for_config(config, only_max_observed_budget=True)
+
+        costs = []
+        for i, r, b in instance_seed_budget_keys:
+            k = RunKey(id_, i, r, b)
+            costs.append(self.data[k].cost)
+        return costs
+
+    def average_cost(
+        self,
+        config: Configuration,
+        instance_seed_budget_keys: typing.Optional[typing.Iterable[InstSeedBudgetKey]] = None,
+    ) -> float:
+        """Return the average cost of a configuration.
+
+        This is the mean of costs of all instance-seed pairs.
+
+        Parameters
+        ----------
+        config : Configuration
+            Configuration to calculate objective for
+        instance_seed_budget_keys : list, optional (default=None)
+            List of tuples of instance-seeds-budget keys. If None, the run_history is
+            queried for all runs of the given configuration.
+
+        Returns
+        ----------
+        Cost: float
+            Average cost
+        """
+        return float(np.mean(self._cost(config, instance_seed_budget_keys)))
+
+    def sum_cost(
+        self,
+        config: Configuration,
+        instance_seed_budget_keys: typing.Optional[typing.Iterable[InstSeedBudgetKey]] = None,
+    ) -> float:
+        """Return the sum of costs of a configuration.
+
+        This is the sum of costs of all instance-seed pairs.
+
+        Parameters
+        ----------
+        config : Configuration
+            Configuration to calculate objective for
+        instance_seed_budget_keys : list, optional (default=None)
+            List of tuples of instance-seeds-budget keys. If None, the run_history is
+            queried for all runs of the given configuration.
+
+        Returns
+        ----------
+        sum_cost: float
+            Sum of costs of config
+        """
+        return float(np.sum(self._cost(config, instance_seed_budget_keys)))
+
+    def compute_all_costs(self, instances: typing.Optional[typing.List[str]] = None) -> None:
+        """Computes the cost of all configurations from scratch and overwrites
+        self.cost_perf_config and self.runs_per_config accordingly;
+
+        Note
+        ----
+        This method is only used for ``merge_foreign_data`` and should be removed.
+
+        Parameters
+        ----------
+        instances: typing.List[str]
+            list of instances; if given, cost is only computed wrt to this instance set
+        """
+        self._cost_per_config = {}
+        self.num_runs_per_config = {}
+        for config, config_id in self.config_ids.items():
+            # removing duplicates while keeping the order
+            inst_seed_budgets = list(dict.fromkeys(self.get_runs_for_config(config, only_max_observed_budget=True)))
+            if instances is not None:
+                inst_seed_budgets = list(
+                    filter(
+                        lambda x: x.instance in typing.cast(typing.List, instances), inst_seed_budgets
+                    )
+                )
+
+            if inst_seed_budgets:  # can be empty if never saw any runs on <instances>
+                perf = self.average_cost(config, inst_seed_budgets)
+                self._cost_per_config[config_id] = perf
+                self.num_runs_per_config[config_id] = len(inst_seed_budgets)
+
+    def get_instance_costs_for_config(self, config: Configuration) -> typing.Dict[str, typing.List[float]]:
+        """ Returns the average cost per instance (across seeds) for a configuration
+
+        If the runhistory contains budgets, only the highest budget for a configuration is returned.
+
+        Note
+        ----
+        This is used by the pSMAC facade to determine the incumbent after the evaluation.
+
+        Parameters
+        ----------
+        config : Configuration from ConfigSpace
+            Parameter configuration
+
+        Returns
+        -------
+        cost_per_inst: dict<instance name<str>, cost<float>>
+        """
+        runs_ = self.get_runs_for_config(config, only_max_observed_budget=True)
+        cost_per_inst = {}  # type: typing.Dict[str, typing.List[float]]
+        for inst, seed, budget in runs_:
+            cost_per_inst[inst] = cost_per_inst.get(inst, [])
+            rkey = RunKey(self.config_ids[config], inst, seed, budget)
+            vkey = self.data[rkey]
+            cost_per_inst[inst].append(vkey.cost)
+        cost_per_inst = dict([(inst, np.mean(costs)) for inst, costs in cost_per_inst.items()])
+        return cost_per_inst
