@@ -1,5 +1,4 @@
 import abc
-from collections import OrderedDict
 import logging
 import typing
 
@@ -89,21 +88,23 @@ class AbstractRunHistory2EPM(object):
         self.scale_perc = scale_perc
 
         # Configuration
-        self.success_states = success_states
         self.impute_censored_data = impute_censored_data
-        self.impute_state = impute_state
-        self.cutoff_time = self.scenario.cutoff
+        self.cutoff_time = self.scenario.cutoff  # type: ignore[attr-defined] # noqa F821
         self.imputor = imputor
 
         # Fill with some default values
         if rng is None:
             self.rng = np.random.RandomState(seed=1)
 
-        if self.impute_state is None:
+        if impute_state is None:
             self.impute_state = [StatusType.CAPPED, ]
+        else:
+            self.impute_state = impute_state
 
-        if self.success_states is None:
+        if success_states is None:
             self.success_states = [StatusType.SUCCESS, ]
+        else:
+            self.success_states = success_states
 
         self.instance_features = scenario.feature_dict
         self.n_feats = scenario.n_features
@@ -130,16 +131,16 @@ class AbstractRunHistory2EPM(object):
                              type(self.imputor))
 
         # Learned statistics
-        self.min_y = None
-        self.max_y = None
-        self.perc = None
+        self.min_y = np.NaN
+        self.max_y = np.NaN
+        self.perc = np.NaN
 
     @abc.abstractmethod
     def _build_matrix(self, run_dict: typing.Mapping[RunKey, RunValue],
                       runhistory: RunHistory,
                       instances: list = None,
                       return_time_as_y: bool = False,
-                      store_statistics: bool = False):
+                      store_statistics: bool = False) -> typing.Tuple[np.ndarray, np.ndarray]:
         """Builds x,y matrixes from selected runs from runhistory
 
         Parameters
@@ -162,7 +163,11 @@ class AbstractRunHistory2EPM(object):
         """
         raise NotImplementedError()
 
-    def transform(self, runhistory: RunHistory):
+    def transform(
+        self,
+        runhistory: RunHistory,
+        budget_subset: typing.Optional[typing.List] = None,
+    ) -> typing.Tuple[np.ndarray, np.ndarray]:
         """Returns vector representation of runhistory; if imputation is
         disabled, censored (TIMEOUT with time < cutoff) will be skipped
 
@@ -170,6 +175,7 @@ class AbstractRunHistory2EPM(object):
         ----------
         runhistory : smac.runhistory.runhistory.RunHistory
             Runhistory containing all evaluated configurations/instances
+        budget_subset : list of budgets to consider
 
         Returns
         -------
@@ -184,6 +190,11 @@ class AbstractRunHistory2EPM(object):
         s_run_dict = {run: runhistory.data[run] for run in runhistory.data.keys()
                       if runhistory.data[run].status in self.success_states}
 
+        # consider only runs on a given budget
+        if budget_subset is not None:
+            s_run_dict = {run: runhistory.data[run] for run in runhistory.data.keys()
+                          if run.budget in budget_subset}
+
         # Store a list of instance IDs
         s_instance_id_list = [k.instance_id for k in s_run_dict.keys()]
         X, Y = self._build_matrix(run_dict=s_run_dict, runhistory=runhistory,
@@ -191,12 +202,12 @@ class AbstractRunHistory2EPM(object):
 
         # Also get TIMEOUT runs
         t_run_dict = {run: runhistory.data[run] for run in runhistory.data.keys()
-                      if runhistory.data[run].status == StatusType.TIMEOUT and
-                      runhistory.data[run].time >= self.cutoff_time}
+                      if runhistory.data[
+                          run].status == StatusType.TIMEOUT and runhistory.data[run].time >= self.cutoff_time}
         t_instance_id_list = [k.instance_id for k in s_run_dict.keys()]
 
         # use penalization (e.g. PAR10) for EPM training
-        store_statistics = True if self.min_y is None else False
+        store_statistics = True if np.isnan(self.min_y) else False
         tX, tY = self._build_matrix(run_dict=t_run_dict, runhistory=runhistory,
                                     instances=t_instance_id_list, store_statistics=store_statistics)
 
@@ -208,8 +219,8 @@ class AbstractRunHistory2EPM(object):
         if self.impute_censored_data:
             # Get all censored runs
             c_run_dict = {run: runhistory.data[run] for run in runhistory.data.keys()
-                          if runhistory.data[run].status in self.impute_state and
-                          runhistory.data[run].time < self.cutoff_time}
+                          if runhistory.data[
+                              run].status in self.impute_state and runhistory.data[run].time < self.cutoff_time}
             if len(c_run_dict) == 0:
                 self.logger.debug("No censored data found, skip imputation")
                 # If we do not impute, we also return TIMEOUT data
@@ -240,6 +251,7 @@ class AbstractRunHistory2EPM(object):
 
                 # return imp_Y in PAR depending on the used threshold in
                 # imputor
+                assert isinstance(self.imputor, BaseImputor)  # please mypy
                 imp_Y = self.imputor.impute(censored_X=cen_X, censored_y=cen_Y,
                                             uncensored_X=X, uncensored_y=Y)
 
@@ -269,7 +281,7 @@ class AbstractRunHistory2EPM(object):
         """
         raise NotImplementedError
 
-    def get_X_y(self, runhistory: RunHistory):
+    def get_X_y(self, runhistory: RunHistory) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:
         """Simple interface to obtain all data in runhistory in X, y format
 
         Parameters
@@ -290,7 +302,7 @@ class AbstractRunHistory2EPM(object):
         y = []
         cen = []
         feature_dict = self.scenario.feature_dict
-        params = self.scenario.cs.get_hyperparameters()
+        params = self.scenario.cs.get_hyperparameters()  # type: ignore[attr-defined] # noqa F821
         for k, v in runhistory.data.items():
             config = runhistory.ids_config[k.config_id]
             x = [config.get(p.name) for p in params]
@@ -310,7 +322,7 @@ class RunHistory2EPM4Cost(AbstractRunHistory2EPM):
                       runhistory: RunHistory,
                       instances: list = None,
                       return_time_as_y: bool = False,
-                      store_statistics: bool = False):
+                      store_statistics: bool = False) -> typing.Tuple[np.ndarray, np.ndarray]:
         """"Builds X,y matrixes from selected runs from runhistory
 
         Parameters
@@ -398,7 +410,6 @@ class RunHistory2EPM4LogCost(RunHistory2EPM4Cost):
         np.ndarray
         """
 
-
         # ensure that minimal value is larger than 0
         if np.any(values <= 0):
             self.logger.warning(
@@ -439,7 +450,7 @@ class RunHistory2EPM4ScaledCost(RunHistory2EPM4Cost):
 
 class RunHistory2EPM4InvScaledCost(RunHistory2EPM4Cost):
     """TODO"""
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs):  # type: ignore[no-untyped-def] # noqa F723
         super().__init__(**kwargs)
         if self.instance_features is not None:
             if len(self.instance_features) > 1:
@@ -473,7 +484,7 @@ class RunHistory2EPM4InvScaledCost(RunHistory2EPM4Cost):
 
 class RunHistory2EPM4SqrtScaledCost(RunHistory2EPM4Cost):
     """TODO"""
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs):  # type: ignore[no-untyped-def]  # noqa F723
         super().__init__(**kwargs)
         if self.instance_features is not None:
             if len(self.instance_features) > 1:
@@ -540,7 +551,7 @@ class RunHistory2EPM4EIPS(AbstractRunHistory2EPM):
     def _build_matrix(self, run_dict: typing.Mapping[RunKey, RunValue],
                       runhistory: RunHistory, instances: typing.List[str] = None,
                       return_time_as_y: bool = False,
-                      store_statistics: bool = False):
+                      store_statistics: bool = False) -> typing.Tuple[np.ndarray, np.ndarray]:
         """TODO"""
         if return_time_as_y:
             raise NotImplementedError()
@@ -570,7 +581,7 @@ class RunHistory2EPM4EIPS(AbstractRunHistory2EPM):
 
         return X, y
 
-    def transform_response_values(self, values: np.ndarray):
+    def transform_response_values(self, values: np.ndarray) -> typing.Tuple[np.ndarray]:
         """Transform function response values.
 
         Transform the runtimes by a log transformation (log(1 + runtime).
